@@ -3,15 +3,15 @@
 =============================
 
 :mod:`darc` is designed as a swiss army knife for darkweb crawling.
-It integrates |requests|_ to collect HTTP request and response
+It integrates :mod:`requests` to collect HTTP request and response
 information, such as cookies, header fields, etc. It also bundles
-|selenium|_ to provide a fully rendered web page and screenshot
+:mod:`selenium` to provide a fully rendered web page and screenshot
 of such view.
 
 As the websites can be sometimes irritating for their anti-robots
 verification, login requirements, etc., the :mod:`darc` project
 also privides hooks to customise crawling behaviours around both
-|requests|_ and |selenium|_.
+:mod:`requests` and :mod:`selenium`.
 
 .. seealso::
 
@@ -27,14 +27,14 @@ and the fallback slowest single-threaded solutions when crawling.
 
 .. note::
 
-   When rendering the target website using |selenium|_ powered by
+   When rendering the target website using :mod:`selenium` powered by
    the renown Google Chrome, it will require much memory to run.
    Thus, the three solutions mentioned above would only toggle the
-   behaviour around the use of |selenium|_.
+   behaviour around the use of :mod:`selenium`.
 
 """
 
 from darc.process import process as darc  # pylint: disable=unused-import
 
 __all__ = ['darc']
-__version__ = '0.4.1'
+__version__ = '0.5.0.post2'
