@@ -5,15 +5,19 @@ import datetime
 import getpass
 import json
 import math
+import multiprocessing
 import os
 import pprint
 import re
 import shutil
 import sys
+import threading
 
 import redis
 import stem.util.term
 
+import darc.typing as typing
+from darc._compat import nullcontext
 from darc.error import render_error
 
 # reboot mode?
@@ -32,11 +36,6 @@ FORCE = bool(int(os.getenv('DARC_FORCE', '0')))
 CHECK_NG = bool(int(os.getenv('DARC_CHECK_CONTENT_TYPE', '0')))
 CHECK = bool(int(os.getenv('DARC_CHECK', '0'))) or CHECK_NG
 
-# save mode?
-_SAVE = bool(int(os.getenv('DARC_SAVE', '0')))
-SAVE_REQUESTS = bool(int(os.getenv('DARC_SAVE_REQUESTS', '0'))) or _SAVE
-SAVE_SELENIUM = bool(int(os.getenv('DARC_SAVE_SELENIUM', '0'))) or _SAVE
-
 # root path
 ROOT = os.path.dirname(os.path.abspath(__file__))
 CWD = os.path.realpath(os.curdir)
@@ -64,8 +63,6 @@ os.makedirs(PATH_MISC, exist_ok=True)
 
 # link file mapping
 PATH_LN = os.path.join(PATH_DB, 'link.csv')
-PATH_QR = os.path.join(PATH_DB, '_queue_requests.txt')
-PATH_QS = os.path.join(PATH_DB, '_queue_selenium.txt')
 
 # PID file
 PATH_ID = os.path.join(PATH_DB, 'darc.pid')
@@ -159,7 +156,7 @@ del _SE_WAIT
 SE_EMPTY = '<html><head></head><body></body></html>'
 
 # Redis client
-REDIS = redis.Redis.from_url(os.getenv('REDIS_URL', 'redis://127.0.0.1'))
+REDIS = redis.Redis.from_url(os.getenv('REDIS_URL', 'redis://localhost'), decode_components=True)
 
 # selenium wait time
 _DARC_WAIT = float(os.getenv('DARC_WAIT', '60'))
@@ -188,3 +185,18 @@ def getpid() -> int:
         with open(PATH_ID) as file:
             return int(file.read().strip())
     return -1
+
+
+def get_lock() -> typing.Union[multiprocessing.Lock, threading.Lock, nullcontext]:
+    """Get a lock.
+
+    Returns:
+        Lock context based on :data:`~darc.const.FLAG_MP`
+        and :data:`~darc.const.FLAG_TH`.
+
+    """
+    if FLAG_MP:
+        return multiprocessing.Lock()
+    if FLAG_TH:
+        return threading.Lock()
+    return nullcontext()
