@@ -4,8 +4,8 @@
 
 The :mod:`darc.crawl` module provides two types of crawlers.
 
-* :func:`~darc.crawl.crawler` -- crawler powered by |requests|_
-* :func:`~darc.crawl.loader` -- crawler powered by |selenium|_
+* :func:`~darc.crawl.crawler` -- crawler powered by :mod:`requests`
+* :func:`~darc.crawl.loader` -- crawler powered by :mod:`selenium`
 
 """
 
@@ -27,52 +27,53 @@ import urllib3
 
 from darc._compat import datetime
 from darc.const import FORCE, SE_EMPTY
-from darc.db import save_requests, save_selenium
-from darc.error import render_error
+from darc.db import (drop_hostname, drop_requests, drop_selenium, have_hostname, save_requests,
+                     save_selenium)
+from darc.error import LinkNoReturn, render_error
 from darc.link import Link
 from darc.parse import (check_robots, extract_links, get_content_type, match_host, match_mime,
                         match_proxy)
-from darc.proxy.bitcoin import save_bitcoin
-from darc.proxy.data import save_data
-from darc.proxy.ed2k import save_ed2k
 from darc.proxy.i2p import fetch_hosts, read_hosts
-from darc.proxy.irc import save_irc
-from darc.proxy.magnet import save_magnet
-from darc.proxy.mail import save_mail
 from darc.proxy.null import fetch_sitemap, save_invalid
 from darc.requests import request_session
-from darc.save import has_folder, sanitise, save_file, save_headers, save_html
+from darc.save import save_headers
 from darc.selenium import request_driver
 from darc.sites import crawler_hook, loader_hook
 from darc.submit import submit_new_host, submit_requests, submit_selenium
 
 
 def crawler(link: Link):
-    """Single |requests|_ crawler for a entry link.
+    """Single :mod:`requests` crawler for a entry link.
 
     Args:
-        link: URL to be crawled by |requests|_.
+        link: URL to be crawled by :mod:`requests`.
 
     The function will first parse the URL using
     :func:`~darc.link.parse_link`, and check if need to crawl the
-    URL (c.f. :data:`~darc.const.PROXY_WHITE_LIST`, :data:`~darc.const.PROXY_BLACK_LIST`
-    , :data:`~darc.const.LINK_WHITE_LIST` and :data:`~darc.const.LINK_BLACK_LIST`);
-    if true, then crawl the URL with |requests|_.
+    URL (c.f. :data:`~darc.const.PROXY_WHITE_LIST`, :data:`~darc.const.PROXY_BLACK_LIST`,
+    :data:`~darc.const.LINK_WHITE_LIST` and :data:`~darc.const.LINK_BLACK_LIST`);
+    if true, then crawl the URL with :mod:`requests`.
 
     If the URL is from a brand new host, :mod:`darc` will first try
     to fetch and save ``robots.txt`` and sitemaps of the host
-    (c.f. :func:`~darc.save.save_robots` and :func:`~darc.save.save_sitemap`),
-    and extract then save the links from sitemaps (c.f. :func:`~darc.parse.read_sitemap`)
+    (c.f. :func:`~darc.proxy.null.save_robots` and :func:`~darc.proxy.null.save_sitemap`),
+    and extract then save the links from sitemaps (c.f. :func:`~darc.proxy.null.read_sitemap`)
     into link database for future crawling (c.f. :func:`~darc.db.save_requests`).
+
+    .. note::
+
+       A host is new if :func:`~darc.db.have_hostname` returns :data:`True`.
+
+       If :func:`darc.proxy.null.fetch_sitemap` and/or :func:`darc.proxy.i2p.fetch_hosts`
+       failed when fetching such documents, the host will be removed from the
+       hostname database through :func:`~darc.db.drop_hostname`, and considered
+       as new when next encounter.
+
     Also, if the submission API is provided, :func:`~darc.submit.submit_new_host`
     will be called and submit the documents just fetched.
 
-    .. seealso::
-
-        * :func:`darc.proxy.null.fetch_sitemap`
-
     If ``robots.txt`` presented, and :data:`~darc.const.FORCE` is
-    ``False``, :mod:`darc` will check if allowed to crawl the URL.
+    :data:`False`, :mod:`darc` will check if allowed to crawl the URL.
 
     .. note::
 
@@ -88,14 +89,17 @@ def crawler(link: Link):
 
         If :exc:`requests.exceptions.InvalidSchema` is raised, the link
         will be saved by :func:`~darc.proxy.null.save_invalid`. Further
-        processing is dropped.
+        processing is dropped, and the link will be removed from the
+        :mod:`requests` database through :func:`~darc.db.drop_requests`.
+
+        If :exc:`~darc.error.LinkNoReturn` is raised, the link will be
+        removed from the :mod:`requests` database through
+        :func:`~darc.db.drop_requests`.
 
     If the content type of response document is not ignored (c.f.
     :data:`~darc.const.MIME_WHITE_LIST` and :data:`~darc.const.MIME_BLACK_LIST`),
-    :mod:`darc` will save the document using :func:`~darc.save.save_html` or
-    :func:`~darc.save.save_file` accordingly. And if the submission API
-    is provided, :func:`~darc.submit.submit_requests` will be called and
-    submit the document just fetched.
+    :func:`~darc.submit.submit_requests` will be called and submit the document
+    just fetched.
 
     If the response document is HTML (``text/html`` and ``application/xhtml+xml``),
     :func:`~darc.parse.extract_links` will be called then to extract
@@ -105,7 +109,7 @@ def crawler(link: Link):
     And if the response status code is between ``400`` and ``600``,
     the URL will be saved back to the link database
     (c.f. :func:`~darc.db.save_requests`). If **NOT**, the URL will
-    be saved into |selenium|_ link database to proceed next steps
+    be saved into :mod:`selenium` link database to proceed next steps
     (c.f. :func:`~darc.db.save_selenium`).
 
     """
@@ -114,54 +118,22 @@ def crawler(link: Link):
         if match_proxy(link.proxy):
             print(render_error(f'[REQUESTS] Ignored proxy type from {link.url} ({link.proxy})',
                                stem.util.term.Color.YELLOW), file=sys.stderr)  # pylint: disable=no-member
-            return
-
-        # save bitcoin address
-        if link.proxy == 'bitcoin':
-            save_bitcoin(link)
-            return
-
-        # save ed2k link
-        if link.proxy == 'ed2k':
-            save_ed2k(link)
-            return
-
-        # save magnet link
-        if link.proxy == 'magnet':
-            save_magnet(link)
-            return
-
-        # save email address
-        if link.proxy == 'mail':
-            save_mail(link)
-            return
-
-        # save IRC address
-        if link.proxy == 'irc':
-            save_irc(link)
-            return
-
-        # save data URI
-        if link.proxy == 'data':
-            try:
-                save_data(link)
-            except ValueError as error:
-                print(render_error(f'[REQUESTS] Failed to save data URI from {link.url} <{error}>',
-                                   stem.util.term.Color.RED), file=sys.stderr)  # pylint: disable=no-member
+            drop_requests(link)
             return
 
         if match_host(link.host):
             print(render_error(f'[REQUESTS] Ignored hostname from {link.url} ({link.proxy})',
                                stem.util.term.Color.YELLOW), file=sys.stderr)  # pylint: disable=no-member
+            drop_requests(link)
             return
 
         # timestamp
         timestamp = datetime.now()
 
         # if it's a new host
-        new_host = has_folder(link) is None
+        if not have_hostname(link):
+            partial = False
 
-        if new_host:
             if link.proxy not in ('zeronet', 'freenet'):
                 # fetch sitemap.xml
                 try:
@@ -169,6 +141,7 @@ def crawler(link: Link):
                 except Exception:
                     error = f'[Error fetching sitemap of {link.url}]' + os.linesep + traceback.format_exc() + '-' * shutil.get_terminal_size().columns  # pylint: disable=line-too-long
                     print(render_error(error, stem.util.term.Color.CYAN), file=sys.stderr)  # pylint: disable=no-member
+                    partial = True
 
             if link.proxy == 'i2p':
                 # fetch hosts.txt
@@ -177,9 +150,12 @@ def crawler(link: Link):
                 except Exception:
                     error = f'[Error subscribing hosts from {link.url}]' + os.linesep + traceback.format_exc() + '-' * shutil.get_terminal_size().columns  # pylint: disable=line-too-long
                     print(render_error(error, stem.util.term.Color.CYAN), file=sys.stderr)  # pylint: disable=no-member
+                    partial = True
 
-            # submit data
-            submit_new_host(timestamp, link)
+            # submit data / drop hostname from db
+            if partial:
+                drop_hostname(link)
+            submit_new_host(timestamp, link, partial=partial)
 
         if not FORCE and not check_robots(link):
             print(render_error(f'[REQUESTS] Robots disallowed link from {link.url}',
@@ -194,11 +170,17 @@ def crawler(link: Link):
                 print(render_error(f'[REQUESTS] Failed on {link.url} <{error}>',
                                    stem.util.term.Color.RED), file=sys.stderr)  # pylint: disable=no-member
                 save_invalid(link)
+                drop_requests(link)
                 return
             except requests.RequestException as error:
                 print(render_error(f'[REQUESTS] Failed on {link.url} <{error}>',
                                    stem.util.term.Color.RED), file=sys.stderr)  # pylint: disable=no-member
-                save_requests(link.url, single=True)
+                save_requests(link, single=True)
+                return
+            except LinkNoReturn:
+                print(render_error(f'[REQUESTS] Removing from database: {link.url}',
+                                   stem.util.term.Color.YELLOW), file=sys.stderr)  # pylint: disable=no-member
+                drop_requests(link)
                 return
 
             # save headers
@@ -210,24 +192,18 @@ def crawler(link: Link):
                 print(render_error(f'[REQUESTS] Generic content type from {link.url} ({ct_type})',
                                    stem.util.term.Color.YELLOW), file=sys.stderr)  # pylint: disable=no-member
 
-                text = response.content
-                try:
-                    path = save_file(timestamp, link, text)
-                except Exception as error:
-                    print(render_error(f'[REQUESTS] Failed to save generic file from {link.url} <{error}>',
-                                       stem.util.term.Color.RED), file=sys.stderr)  # pylint: disable=no-member
-                    return
-
                 # probably hosts.txt
                 if link.proxy == 'i2p' and ct_type in ['text/plain', 'text/text']:
-                    with open(path) as hosts_file:
-                        save_requests(read_hosts(hosts_file))
+                    text = response.text
+                    save_requests(read_hosts(text))
 
                 if match_mime(ct_type):
+                    drop_requests(link)
                     return
 
                 # submit data
-                submit_requests(timestamp, link, response, session)
+                data = response.content
+                submit_requests(timestamp, link, response, session, data, html=False)
 
                 return
 
@@ -238,14 +214,11 @@ def crawler(link: Link):
                 save_requests(link.url, single=True)
                 return
 
-            # save HTML
-            save_html(timestamp, link, html, raw=True)
-
             # submit data
-            submit_requests(timestamp, link, response, session)
+            submit_requests(timestamp, link, response, session, html, html=True)
 
             # add link to queue
-            save_requests(extract_links(link, html), score=0)
+            save_requests(extract_links(link, html), score=0, nx=True)
 
             if not response.ok:
                 print(render_error(f'[REQUESTS] Failed on {link.url} [{response.status_code}]',
@@ -255,7 +228,6 @@ def crawler(link: Link):
 
             # add link to queue
             save_selenium(link, single=True, score=0, nx=True)
-            save_requests(link, single=True)
     except Exception:
         error = f'[Error from {link.url}]' + os.linesep + traceback.format_exc() + '-' * shutil.get_terminal_size().columns  # pylint: disable=line-too-long
         print(render_error(error, stem.util.term.Color.CYAN), file=sys.stderr)  # pylint: disable=no-member
@@ -264,21 +236,26 @@ def crawler(link: Link):
 
 
 def loader(link: Link):
-    """Single |selenium|_ loader for a entry link.
+    """Single :mod:`selenium` loader for a entry link.
 
     Args:
-        Link: URL to be crawled by |requests|_.
+        Link: URL to be crawled by :mod:`selenium`.
 
     The function will first parse the URL using :func:`~darc.link.parse_link`
-    and start loading the URL using |selenium|_ with Google Chrome.
+    and start loading the URL using :mod:`selenium` with Google Chrome.
 
     At this point, :mod:`darc` will call the customised hook function
     from :mod:`darc.sites` to load and return the original
-    |Chrome|_ object.
+    :class:`selenium.webdriver.Chrome` object.
 
-    If successful, the rendered source HTML document will be saved
-    using :func:`~darc.save.save_html`, and a full-page screenshot
-    will be taken and saved.
+    .. note::
+
+        If :exc:`~darc.error.LinkNoReturn` is raised, the link will be
+        removed from the :mod:`selenium` database through
+        :func:`~darc.db.drop_selenium`.
+
+    If successful, the rendered source HTML document will be saved, and a
+    full-page screenshot will be taken and saved.
 
     .. note::
 
@@ -287,21 +264,21 @@ def loader(link: Link):
        height of web page. If the page height is *less than* **1,000 pixels**,
        then :mod:`darc` will by default set the height as **1,000 pixels**.
 
-       Later :mod:`darc` will tell |selenium|_ to resize the window (in
+       Later :mod:`darc` will tell :mod:`selenium` to resize the window (in
        *headless* mode) to **1,024 pixels** in width and **110%** of the
        page height in height, and take a *PNG* screenshot.
 
-    .. seealso::
-
-       * :data:`darc.const.SE_EMPTY`
-       * :data:`darc.const.SE_WAIT`
-
     If the submission API is provided, :func:`~darc.submit.submit_selenium`
     will be called and submit the document just loaded.
 
     Later, :func:`~darc.parse.extract_links` will be called then to
     extract all possible links from the HTML document and save such
-    links into the |requests|_ database (c.f. :func:`~darc.db.save_requests`).
+    links into the :mod:`requests` database (c.f. :func:`~darc.db.save_requests`).
+
+    .. seealso::
+
+       * :data:`darc.const.SE_EMPTY`
+       * :data:`darc.const.SE_WAIT`
 
     """
     print(f'[SELENIUM] Loading {link.url}')
@@ -324,6 +301,11 @@ def loader(link: Link):
                                    stem.util.term.Color.RED), file=sys.stderr)  # pylint: disable=no-member
                 save_selenium(link, single=True)
                 return
+            except LinkNoReturn:
+                print(render_error(f'[SELENIUM] Removing from database: {link.url}',
+                                   stem.util.term.Color.YELLOW), file=sys.stderr)  # pylint: disable=no-member
+                drop_selenium(link)
+                return
 
             # get HTML source
             html = driver.page_source
@@ -334,9 +316,7 @@ def loader(link: Link):
                 save_selenium(link, single=True)
                 return
 
-            # save HTML
-            save_html(timestamp, link, html)
-
+            screenshot = None
             try:
                 # get maximum height
                 height = driver.execute_script('return document.body.scrollHeight')
@@ -347,18 +327,16 @@ def loader(link: Link):
                 driver.set_window_size(1024, math.ceil(height * 1.1))
 
                 # take a full page screenshot
-                path = sanitise(link, timestamp, screenshot=True)
-                driver.save_screenshot(path)
+                screenshot = driver.get_screenshot_as_base64()
             except Exception as error:
                 print(render_error(f'[SELENIUM] Fail to save screenshot from {link.url} <{error}>',
                                    stem.util.term.Color.RED), file=sys.stderr)  # pylint: disable=no-member
 
             # submit data
-            submit_selenium(timestamp, link)
+            submit_selenium(timestamp, link, html, screenshot)
 
             # add link to queue
             save_requests(extract_links(link, html), score=0, nx=True)
-            save_selenium(link, single=True)
     except Exception:
         error = f'[Error from {link.url}]' + os.linesep + traceback.format_exc() + '-' * shutil.get_terminal_size().columns  # pylint: disable=line-too-long
         print(render_error(error, stem.util.term.Color.CYAN), file=sys.stderr)  # pylint: disable=no-member
